package com.marcpg.bbm.factions;

import com.mojang.brigadier.arguments.DoubleArgumentType;
import com.mojang.brigadier.arguments.StringArgumentType;
import com.mojang.brigadier.builder.LiteralArgumentBuilder;
import com.mojang.brigadier.builder.RequiredArgumentBuilder;
import com.mojang.brigadier.tree.LiteralCommandNode;
import io.papermc.paper.command.brigadier.CommandSourceStack;
import io.papermc.paper.command.brigadier.argument.ArgumentTypes;
import io.papermc.paper.command.brigadier.argument.resolvers.selector.PlayerSelectorArgumentResolver;
import net.kyori.adventure.text.Component;
import net.kyori.adventure.text.format.NamedTextColor;
import net.milkbowl.vault2.economy.EconomyResponse;
import org.bukkit.Bukkit;
import org.bukkit.OfflinePlayer;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.entity.PlayerDeathEvent;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.math.BigDecimal;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;
import java.util.*;

public class BountySystem implements Listener {
    public record Bounty(UUID on, UUID by, double reward) {}

    public static final Path FILE = BBMFactions.PLUGIN.getDataPath().resolve("bounties");
    public static final List<Bounty> BOUNTIES = new ArrayList<>();

    public static List<Bounty> getBountiesOn(UUID on) {
        return BOUNTIES.parallelStream().filter(b -> b.on == on).toList();
    }

    public static List<Bounty> getBountiesBy(UUID by) {
        return BOUNTIES.parallelStream().filter(b -> b.by == by).toList();
    }

    public static void load() {
        try {
            if (FILE.toFile().createNewFile()) return;
            Files.readAllLines(FILE).parallelStream()
                    .map(l -> l.split(","))
                    .forEach(b -> BOUNTIES.add(new Bounty(UUID.fromString(b[0]), UUID.fromString(b[1]), Double.parseDouble(b[2]))));
        } catch (IOException e) {
            BBMFactions.LOG.warn("Could not load bounties from file.", e);
        }
    }

    public static void save() {
        try {
            List<String> lines = BOUNTIES.parallelStream()
                    .map(b -> b.on + "|" + b.by + "|" + b.reward)
                    .toList();
            Files.write(FILE, lines, StandardOpenOption.CREATE, StandardOpenOption.WRITE, StandardOpenOption.TRUNCATE_EXISTING);
        } catch (IOException e) {
            BBMFactions.LOG.warn("Could not save bounties to file.", e);
        }
    }

    @EventHandler(ignoreCancelled = true)
    public void onPlayerDeath(@NotNull PlayerDeathEvent event) {
        Player player = event.getPlayer();
        if (player.getKiller() instanceof Player killer && killer.hasPermission("factions.claim-bounty")) {
            List<Bounty> bounties = getBountiesOn(player.getUniqueId());
            if (bounties.isEmpty()) return;

            BigDecimal totalMoney = BigDecimal.valueOf(bounties.stream().mapToDouble(m -> m.reward).sum());

            EconomyResponse response = BBMFactions.ECONOMY.deposit("BBMFactions", player.getUniqueId(), totalMoney);
            if (response.transactionSuccess()) {
                if (bounties.size() == 1) {
                    event.deathMessage(Component.text(killer.getName(), NamedTextColor.WHITE)
                            .append(Component.text(" claimed the ", NamedTextColor.YELLOW))
                            .append(Component.text(BBMFactions.ECONOMY.format("BBMFactions", totalMoney), NamedTextColor.WHITE))
                            .append(Component.text(" bounty on ", NamedTextColor.YELLOW))
                            .append(Component.text(player.getName(), NamedTextColor.WHITE))
                            .append(Component.text(".", NamedTextColor.YELLOW)));
                } else {
                    event.deathMessage(Component.text(killer.getName(), NamedTextColor.WHITE)
                            .append(Component.text(" claimed ", NamedTextColor.YELLOW))
                            .append(Component.text(bounties.size(), NamedTextColor.WHITE))
                            .append(Component.text(" bounties on ", NamedTextColor.YELLOW))
                            .append(Component.text(player.getName(), NamedTextColor.WHITE))
                            .append(Component.text(", totalling ", NamedTextColor.YELLOW))
                            .append(Component.text(BBMFactions.ECONOMY.format("BBMFactions", totalMoney), NamedTextColor.WHITE))
                            .append(Component.text(".", NamedTextColor.YELLOW)));
                }

                for (Bounty b : bounties) {
                    BBMFactions.ECONOMY.withdraw("BBMFactions", b.by, BigDecimal.valueOf(b.reward));
                    BOUNTIES.remove(b);
                }
            }
        }
    }

    @SuppressWarnings("UnstableApiUsage")
    public static LiteralCommandNode<CommandSourceStack> bountyCommand() {
        return LiteralArgumentBuilder.<CommandSourceStack>literal("bounty")
                .then(LiteralArgumentBuilder.<CommandSourceStack>literal("create")
                        .requires(context -> context.getSender() instanceof Player)
                        .then(RequiredArgumentBuilder.<CommandSourceStack, String>argument("player", StringArgumentType.word())
                                .then(RequiredArgumentBuilder.<CommandSourceStack, Double>argument("reward", DoubleArgumentType.doubleArg(5.0))
                                        .executes(context -> {
                                            Player player = (Player) context.getSource();
                                            OfflinePlayer target = Bukkit.getOfflinePlayer(context.getArgument("player", String.class));

                                            Bounty bounty = new Bounty(target.getUniqueId(), player.getUniqueId(), context.getArgument("reward", Double.class));
                                            BOUNTIES.add(bounty);

                                            Bukkit.getServer().broadcast(Component.text(player.getName(), NamedTextColor.WHITE)
                                                    .append(Component.text(" set a bounty on ", NamedTextColor.YELLOW))
                                                    .append(Component.text(Objects.requireNonNullElse(target.getName(), "[Name not found]"), NamedTextColor.WHITE))
                                                    .append(Component.text("!\nFirst to kill them will receive ", NamedTextColor.YELLOW))
                                                    .append(Component.text(BBMFactions.ECONOMY.format("BBMFactions", BigDecimal.valueOf(bounty.reward)), NamedTextColor.WHITE))
                                                    .append(Component.text(".", NamedTextColor.YELLOW)));
                                            return 1;
                                        })
                                )
                        )
                )
                .then(LiteralArgumentBuilder.<CommandSourceStack>literal("withdraw")
                        .requires(context -> context.getSender() instanceof Player)
                        .then(RequiredArgumentBuilder.<CommandSourceStack, String>argument("player", StringArgumentType.word())
                                .suggests((context, builder) -> {
                                    getBountiesBy(((Player) context.getSource().getSender()).getUniqueId())
                                            .forEach(b -> builder.suggest(Bukkit.getOfflinePlayer(b.on).getName()));
                                    return builder.buildFuture();
                                })
                                .executes(context -> {
                                    UUID by = ((Player) context.getSource().getSender()).getUniqueId();
                                    OfflinePlayer on = Bukkit.getOfflinePlayer(context.getArgument("player", String.class));
                                    if (BOUNTIES.removeIf(bounty -> bounty.by == by && bounty.on == on.getUniqueId())) {
                                        context.getSource().getSender().sendMessage(Component.text("Successfully withdrawn your bounty on " + on.getName() + ".", NamedTextColor.YELLOW));
                                    } else {
                                        context.getSource().getSender().sendMessage(Component.text("You did not set a bounty on " + on.getName() + ", or it was already claimed.", NamedTextColor.YELLOW));
                                    }
                                    return 1;
                                })
                        )
                )
                .then(LiteralArgumentBuilder.<CommandSourceStack>literal("top")
                        .executes(context -> {
                            CommandSender source = context.getSource().getSender();
                            source.sendMessage(Component.text("Top 10 highest bounties:", NamedTextColor.YELLOW));
                            List<Bounty> top10Bounties = BOUNTIES.stream().sorted(Comparator.comparingDouble(b -> -b.reward)).limit(10).toList();
                            for (int i = 0; i < top10Bounties.size(); i++) {
                                source.sendMessage(Component.text((i + 1) + ". " + Bukkit.getOfflinePlayer(top10Bounties.get(i).on).getName() + " (" + BBMFactions.ECONOMY.format("BBMFactions", BigDecimal.valueOf(-top10Bounties.get(i).reward)) + ")", NamedTextColor.YELLOW));
                            }
                            return 1;
                        })
                )
                .then(LiteralArgumentBuilder.<CommandSourceStack>literal("check")
                        .requires(context -> context.getSender() instanceof Player)
                        .then(RequiredArgumentBuilder.<CommandSourceStack, PlayerSelectorArgumentResolver>argument("player", ArgumentTypes.player())
                                .executes(context -> {
                                    OfflinePlayer target = Bukkit.getOfflinePlayer(context.getArgument("player", String.class));
                                    List<Bounty> bounties = getBountiesOn(target.getUniqueId());
                                    if (bounties.isEmpty()) {
                                        context.getSource().getSender().sendMessage(Component.text("There are no bounties on " + target.getName() + ".", NamedTextColor.YELLOW));
                                    } else if (bounties.size() == 1) {
                                        context.getSource().getSender().sendMessage(Component.text("There is one bounty on " + target.getName() + ", rewarding " + BBMFactions.ECONOMY.format("BBMFactions", BigDecimal.valueOf(bounties.getFirst().reward)) + ".", NamedTextColor.YELLOW));
                                    } else {
                                        BigDecimal totalMoney = BigDecimal.valueOf(bounties.stream().mapToDouble(m -> m.reward).sum());
                                        context.getSource().getSender().sendMessage(Component.text("There are " + bounties.size() + " bounties on " + target.getName() + ", rewarding " + BBMFactions.ECONOMY.format("BBMFactions", totalMoney) + ".", NamedTextColor.YELLOW));
                                    }
                                    return 1;
                                })
                        )
                )
                .build();
    }
}
